/* You may find the license in the LICENSE file */

const DTA = {_dh: {}};
Components.utils.import("resource://dta/api.jsm", DTA);
Components.utils.import("resource://dta/support/downloadHelper.jsm", DTA._dh);


function DTA_showPreferences() {
	var instantApply = DTA.Preferences.get("browser.preferences.instantApply", false);
	window.openDialog(
		'chrome://dta/content/preferences/prefs.xul',
		'dtaPrefs',
		'chrome,titlebar,toolbar,resizable,centerscreen'+ (instantApply ? ',dialog=no' : '')
	);
}

function DTA_DropProcessor(func, multiple) {
	this.func = func;
	if (multiple) {
		this.canHandleMultipleItems = true;
	}
};
DTA_DropProcessor.prototype = {
	getSupportedFlavours: function() {
		if (!this._flavors) {
			this._flavors = new FlavourSet();
			this._flavors.appendFlavour('text/x-moz-url');
		}	
		return this._flavors;
	},
	onDragOver: function() {},
	onDrop: function (evt, dropdata, session) {
		if (!dropdata) {
			return;
		}
		let url = null;
		try {
			url = transferUtils.retrieveURLFromData(dropdata.data, dropdata.flavour.contentType);
			if (!DTA.isLinkOpenable(url)) {
				throw new Components.Exception("Link cannot be opened!");
			}
			url = DTA.IOService.newURI(url, null, null);
		}
		catch (ex) {
			DTA.Debug.log("Failed to process drop", ex);
			return;
		}
		let doc = document.commandDispatcher.focusedWindow.document;
		let ref = doc ? DTA.getRef(doc) : null;		
		
		if (url) {
			url = new DTA.URL(DTA.getLinkPrintMetalink(url) || url);
			this.func(url, ref);			
		}
	}
};

var DTA_DropTDTA = new DTA_DropProcessor(function(url, ref) { DTA.saveSingleLink(window, true, url, ref); });
var DTA_DropDTA = new DTA_DropProcessor(function(url, ref) { DTA.saveSingleLink(window, false, url, ref); });

this.__defineGetter__('DTA_Mediator', function() {
	delete this.DTA_Mediator;
	this.DTA_Mediator = {
		open: function DTA_Mediator_open(url, ref) {
			this.openUrl(window, url, ref);
		}
	};
	Components.utils.import('resource://dta/support/mediator.jsm', this.DTA_Mediator);
	return this.DTA_Mediator;
});

/* Compat; mostly FlashGot, maybe others */
// Obsolete; will be removed in 2.++ timeframe
const DTA_AddingFunctions = {
	ios: DTA.IOService,
	composeURL: function() DTA.composeURL.apply(this, arguments),
	applyWithWindow: function(func, args) {
		DTA.Debug.logString("Obsolete function called: " + func.name);
		let args = Array.map(args, function(e) e);
		args[0] = window;
		return func.apply(DTA, args);
	},
	saveSingleLink: function() this.applyWithWindow(DTA.saveSingleLink, arguments),
	saveLinkArray: function() this.applyWithWindow(DTA.saveLinkArray, arguments),
	turboSaveLinkArray: function() this.applyWithWindow(DTA.turboSaveLinkArray, arguments),
	sendToDown: function() this.applyWithWindow(DTA.sendToDown, arguments),
	turboSendToDown: function() this.applyWithWindow(DTA.turboSendToDown, arguments)
};
const DTA_getLinkPrintMetalink = DTA.getLinkPrintMetalink;
const DTA_URL = DTA.URL;